﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;
using Newtonsoft.Json;
using Pixiview.Illust;
using Xamarin.Essentials;
using Xamarin.Forms;

namespace Pixiview.Utils
{
    public static class Stores
    {
        public static readonly string PersonalFolder = Environment.GetFolderPath(Environment.SpecialFolder.Personal);
        public static readonly string CacheFolder = Environment.GetFolderPath(Environment.SpecialFolder.InternetCache);

        private const string favoriteFile = "favorites.json";
        private const string imageFolder = "img-original";
        private const string previewFolder = "img-master";
        private const string illustFile = "illust.json";

        private const string pagesFolder = "pages";
        private const string preloadsFolder = "preloads";
        private const string thumbFolder = "img-thumb";
        private const string userFolder = "user-profile";

        private static readonly object sync = new object();

        public static bool NetworkAvailable
        {
            get
            {
                try
                {
                    return Connectivity.NetworkAccess == NetworkAccess.Internet;
                }
                catch
                {
                    return false;
                }
            }
        }

        public static List<IllustItem> Favorites => GetFavoriteObject().Illusts;
        public static string FavoritesPath => Path.Combine(PersonalFolder, favoriteFile);

        private static IllustFavorite favoriteObject;

        public static IllustFavorite GetFavoriteObject(bool force = false)
        {
            lock (sync)
            {
                if (force || favoriteObject == null)
                {
                    var favorites = LoadFavoritesIllusts();
                    if (favorites != null)
                    {
                        favoriteObject = favorites;
                    }
                    else
                    {
                        favoriteObject = new IllustFavorite
                        {
                            Illusts = new List<IllustItem>()
                        };
                    }
                }
                return favoriteObject;
            }
        }

        public static IllustFavorite LoadFavoritesIllusts(string file = null)
        {
            if (file == null)
            {
                file = FavoritesPath;
            }
            lock (sync)
            {
                return ReadObject<IllustFavorite>(file);
            }
        }

        public static void SaveFavoritesIllusts()
        {
            var file = FavoritesPath;
            lock (sync)
            {
                var data = GetFavoriteObject();
                data.LastFavoriteUtc = DateTime.UtcNow;
                WriteObject(file, data);
            }
        }

        private static T LoadObject<T>(string file, string url, string referer, bool force = false, Func<string, string> action = null)
        {
            string content = null;
            if (!force && file != null && File.Exists(file))
            {
                try
                {
                    content = File.ReadAllText(file);
                }
                catch (Exception ex)
                {
                    App.DebugError("load", $"failed to read file: {file}, error: {ex.Message}");
                }
            }
            if (content == null)
            {
                var response = Download(url, headers =>
                {
                    if (referer != null)
                    {
                        headers.Referrer = new Uri(referer);
                    }
                    headers.Add("user-agent", Configs.UserAgent);
                    headers.Add("accept", Configs.AcceptJson);
                    headers.Add("cookie", Configs.Cookie);
                    headers.Add("x-user-id", Configs.UserId);
                });
                if (response == null)
                {
                    return default;
                }
                using (response)
                {
                    try
                    {
                        content = response.Content.ReadAsStringAsync().Result;
                        if (action != null)
                        {
                            content = action(content);
                        }
                    }
                    catch (Exception ex)
                    {
                        App.DebugError("load.strea", $"failed to read stream, error: {ex.Message}");
                        return default;
                    }

                    if (file != null)
                    {
                        try
                        {
                            var folder = Path.GetDirectoryName(file);
                            if (!Directory.Exists(folder))
                            {
                                Directory.CreateDirectory(folder);
                            }
                            File.WriteAllText(file, content, Encoding.UTF8);
                        }
                        catch (Exception ex)
                        {
                            App.DebugError("save", $"failed to save illust JSON object, error: {ex.Message}");
                        }
                    }
                }
            }
            try
            {
                return JsonConvert.DeserializeObject<T>(content);
            }
            catch (Exception ex)
            {
                App.DebugError("load", $"failed to parse illust JSON object, error: {ex.Message}");
                return default;
            }
        }

        private static T ReadObject<T>(string file)
        {
            string content = null;
            if (File.Exists(file))
            {
                try
                {
                    content = File.ReadAllText(file);
                }
                catch (Exception ex)
                {
                    App.DebugError("read", $"failed to read file: {file}, error: {ex.Message}");
                }
            }
            else
            {
                App.DebugError("read", $"file not found: {file}");
                return default;
            }
            try
            {
                return JsonConvert.DeserializeObject<T>(content);
            }
            catch (Exception ex)
            {
                App.DebugError("read", $"failed to parse illust JSON object, error: {ex.Message}");
                return default;
            }
        }

        private static void WriteObject(string file, object obj)
        {
            var dir = Path.GetDirectoryName(file);
            if (!Directory.Exists(dir))
            {
                Directory.CreateDirectory(dir);
            }
            string content;
            try
            {
                content = JsonConvert.SerializeObject(obj, Formatting.None);
            }
            catch (Exception ex)
            {
                App.DebugError("write", $"failed to serialize object, error: {ex.Message}");
                return;
            }

            try
            {
                File.WriteAllText(file, content, Encoding.UTF8);
            }
            catch (Exception ex)
            {
                App.DebugError("write", $"failed to write file: {file}, error: {ex.Message}");
            }
        }

        public static IllustData LoadIllustData(bool force = false)
        {
            var file = Path.Combine(PersonalFolder, illustFile);
            var result = LoadObject<IllustData>(
                file,
                Configs.UrlIllustList,
                Configs.Referer,
                force: force);
            if (result == null || result.error)
            {
                App.DebugPrint($"error when load illust data: {result?.message} ({force})");
            }
            return result;
        }

        public static IllustPreloadBody LoadIllustPreloadData(string id, bool force = false)
        {
            var file = Path.Combine(CacheFolder, preloadsFolder, $"{id}.json");
            var result = LoadObject<IllustPreloadBody>(
                file,
                string.Format(Configs.UrlIllust, id),
                null,
                force: force,
                action: content =>
                {
                    var index = content.IndexOf(Configs.SuffixPreload);
                    if (index > 0)
                    {
                        index += Configs.SuffixPreloadLength;
                        var end = content.IndexOf('\'', index);
                        if (end > index)
                        {
                            content = content.Substring(index, end - index);
                        }
                    }
                    return content;
                });
            return result;
        }

        public static IllustPageData LoadIllustPageData(string id, bool force = false)
        {
            var file = Path.Combine(CacheFolder, pagesFolder, $"{id}.json");
            var result = LoadObject<IllustPageData>(
                file,
                string.Format(Configs.UrlIllustPage, id),
                string.Format(Configs.UrlIllust, id),
                force: force);
            if (result == null || result.error)
            {
                App.DebugPrint($"error when load page data: {result?.message} ({force})");
            }
            return result;
        }

        public static IllustUserData LoadIllustUserData(string userId, bool force = false)
        {
            var list = LoadObject<IllustUserListData>(
                null,
                string.Format(Configs.UrlIllustUserAll, userId),
                string.Format(Configs.UrlIllustUser, userId),
                force: force);
            if (list == null || list.error)
            {
                App.DebugPrint($"error when load user data: {list?.message} ({force})");
            }

            // TODO
            var ids = string.Join("", list.body.illusts.Keys.Take(20).Select(id => $"ids%5B%5D={id}&"));

            var result = LoadObject<IllustUserData>(
                null,
                string.Format(Configs.UrlIllustUserArtworks, userId, ids, 1),
                string.Format(Configs.UrlIllustUser, userId),
                force: force);
            if (result == null || result.error)
            {
                App.DebugPrint($"error when load user illust data: {result?.message} ({force})");
            }
            return result;
        }

        public static ImageSource LoadIllustImage(string url)
        {
            return LoadImage(url, PersonalFolder, imageFolder);
        }

        public static ImageSource LoadPreviewImage(string url, bool downloading = true)
        {
            return LoadImage(url, PersonalFolder, previewFolder, downloading);
        }

        public static ImageSource LoadThumbnailImage(string url)
        {
            return LoadImage(url, CacheFolder, thumbFolder);
        }

        public static ImageSource LoadUserProfileImage(string url)
        {
            return LoadImage(url, CacheFolder, userFolder);
        }

        public static bool CheckIllustImage(string url)
        {
            var file = Path.Combine(PersonalFolder, imageFolder, Path.GetFileName(url));
            return File.Exists(file);
        }

        public static string GetPreviewImagePath(string url)
        {
            var file = Path.Combine(PersonalFolder, previewFolder, Path.GetFileName(url));
            if (File.Exists(file))
            {
                return file;
            }
            return null;
        }

        private static ImageSource LoadImage(string url, string working, string folder, bool downloading = true)
        {
            var file = Path.Combine(working, folder, Path.GetFileName(url));
            ImageSource image;
            if (File.Exists(file))
            {
                try
                {
                    image = ImageSource.FromFile(file);
                }
                catch (Exception ex)
                {
                    App.DebugError("image.load", $"failed to load image from file: {file}, error: {ex.Message}");
                    image = null;
                }
            }
            else
            {
                image = null;
            }
            if (downloading && image == null)
            {
                file = DownloadImage(url, working, folder);
                if (file != null)
                {
                    return ImageSource.FromFile(file);
                }
            }
            return image;
        }

        private static string DownloadImage(string url, string working, string folder)
        {
            try
            {
                var directory = Path.Combine(working, folder);
                if (!Directory.Exists(directory))
                {
                    Directory.CreateDirectory(directory);
                }
                var file = Path.Combine(directory, Path.GetFileName(url));
                App.DebugPrint($"download, url: {url}");
                var response = Download(url, headers =>
                {
                    headers.Referrer = new Uri(Configs.Referer);
                    headers.Add("user-agent", Configs.UserAgent);
                    headers.Add("accept", Configs.AcceptImage);
                });
                if (response == null)
                {
                    return null;
                }
                using (response)
                using (var fs = File.OpenWrite(file))
                {
                    response.Content.CopyToAsync(fs).Wait();
                    //if (response.Headers.Date != null)
                    //{
                    //    File.SetLastWriteTimeUtc(file, response.Headers.Date.Value.UtcDateTime);
                    //}
                }
                return file;
            }
            catch (Exception ex)
            {
                App.DebugError("image.download", ex.Message);
                return null;
            }
        }

        private static T TryCount<T>(Func<T> func, int tryCount = 3)
        {
            int tries = 0;
            while (tries < tryCount)
            {
                try
                {
                    return func();
                }
                catch (Exception ex)
                {
                    tries++;
                    System.Threading.Thread.Sleep(400);
                    App.DebugError("try.do", $"tries: {tries}, error: {ex.Message}");
                }
            }
            return default;
        }

        private static HttpResponseMessage Download(string url, Action<HttpRequestHeaders> headerAction)
        {
            App.DebugPrint($"GET: {url}");
            var uri = new Uri(url);
            var proxy = Configs.Proxy;
            var handler = new HttpClientHandler
            {
                AutomaticDecompression = DecompressionMethods.GZip | DecompressionMethods.Deflate
            };
            if (proxy != null)
            {
                handler.Proxy = proxy;
                handler.UseProxy = true;
            }
            var client = new HttpClient(handler)
            {
                BaseAddress = new Uri($"{uri.Scheme}://{uri.Host}"),
                Timeout = TimeSpan.FromSeconds(30)
            };
            return TryCount(() =>
            {
                using (var request = new HttpRequestMessage(HttpMethod.Get, uri.PathAndQuery)
                {
                    Version = new Version(2, 0),
                })
                {
                    var headers = request.Headers;
                    headerAction(headers);
                    headers.Add("accept-language", Configs.AcceptLanguage);
                    headers.Add("accept-encoding", Configs.AcceptEncoding);
                    return client.SendAsync(request, HttpCompletionOption.ResponseHeadersRead).Result;
                }
            });
        }
    }

    public static class Configs
    {
        public const string IsProxiedKey = "isProxied";
        public const string HostKey = "host";
        public const string PortKey = "port";

        public static WebProxy Proxy;

        public const int MaxThreads = 3;
        public const string Referer = "https://www.pixiv.net/";
        public const string UrlIllustList = "https://www.pixiv.net/ajax/top/illust?mode=all&lang=zh";
        public const string UrlIllust = "https://www.pixiv.net/artworks/{0}";
        public const string SuffixPreload = " id=\"meta-preload-data\" content='";
        public const int SuffixPreloadLength = 33;  // SuffixPreload.Length
        public const string UrlIllustUserAll = "https://www.pixiv.net/ajax/user/{0}/profile/all?lang=zh";
        public const string UrlIllustUserArtworks = "https://www.pixiv.net/ajax/user/{0}/profile/illusts?{1}work_category=illustManga&is_first_page={2}&lang=zh";
        public const string UrlIllustUser = "https://www.pixiv.net/users/{0}/artworks";
        public const string UrlIllustPage = "https://www.pixiv.net/ajax/illust/{0}/pages?lang=zh";
        public const string UserAgent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.138 Safari/537.36";
        public const string AcceptImage = "image/png,image/svg+xml,image/*;q=0.8,video/*;q=0.8,*/*;q=0.5";
        public const string AcceptJson = "application/json";
        public const string AcceptEncoding = "gzip, deflate";
        public const string AcceptLanguage = "zh-cn";

        public const string UserId = "2603358";
        public const string Cookie =
            "PHPSESSID=2603358_VHyGPeRaz7LpeoFkRsHvjXIpApCMb56a; " +
            "a_type=0; b_type=1; c_type=31; d_type=2; " +
            "p_ab_id=2; p_ab_id_2=6; p_ab_d_id=1155161977; " +
            "privacy_policy_agreement=2; " +
            "login_ever=yes; " +
            "__cfduid=d9fa2d4d1ddd30db85ebb519f9855d2561587806747; " +
            "first_visit_datetime_pc=2019-10-29+22%3A05%3A30; " +
            "yuid_b=NgcXQWQ";

        public static string GetThumbnailUrl(string url)
        {
            url = url.Replace("/custom-thumb/", "/img-master/");
            var index = url.LastIndexOf("_square1200.jpg");
            if (index < 0)
            {
                index = url.LastIndexOf("_custom1200.jpg");
            }
            if (index >= 0)
            {
                return url.Substring(0, index) + "_master1200.jpg";
            }
            return url;
        }
    }

    public static class Routes
    {
        public const string Illust = "illust";
        public const string Detail = "detail";
        public const string Follow = "follow";
        public const string Recommends = "recommends";
        public const string ByUser = "byuser";
        public const string Ranking = "ranking";
        public const string Favorites = "favorites";
        public const string Option = "option";
    }
}
